import sys
from pydub.playback import play
from pythainlp.transliterate import pronunciate
from pythainlp.transliterate import transliterate as thai
import phonemizer
from eng_to_ipa import transcribe as engipa
from nltk.tokenize import SyllableTokenizer
from langdetect import detect
from langdetect.lang_detect_exception import LangDetectException
import pyphen
from flask import Flask, request, jsonify

ipa_hierarchy = [
                "iyɨʉɯuɪʏʊeøɘɵɤoe̞ø̞əɤ̞o̞ɛœɜɞʌɔæɐaɶäɑɒ",  # vowels.
                "m̥mɱn̼n̥nɳ̊ɳɲ̊ɲŋ̊ŋɴ",  # nasals.
                "szʃʒʂʐɕʑɸβfvθ̼ð̼θðθ̠ð̠ɹ̠̊˔ɹ̠˔ɻ̊˔ɻ˔çʝxɣχʁħʕhɦɬɮꞎɭ˔𝼆ʎ̝𝼄ʟ̝",  # fricatives.
                "pbtdkq",  # stops.
            ]
SSP = SyllableTokenizer(sonority_hierarchy=ipa_hierarchy)
ipa_nc = {"ɑ":"ah", "ɑɹ":"ar", "æ":"a", "ɔ":"aw", "ɛəɹ":"air", "eɪ":"ay", "aɪ":"y", "i":"ee", "iː":"ee", "ɪəɹ":"eer", "ɚ":"er", "əɹ":"er", "ɛ":"e","eː":"ay", "oː":"oh", "aː":"ah", "ʌ":"uh", "ə":"uh", "oʊ":"oh", "ɔɹ":"or", "ɔɪ":"oy", "aʊ":"ow", "ʊ":"uu", "uː":"ew", "juː":"yew", "u":"oo", "t᷂͡ʃ":"ch", "t":"t", "g":"g", "ʤ":"j", "d͡ʒ":"j", "ð":"dh", "z":"z", "ʒ":"zh", "θ":"th", "s":"s", "ʃ":"sh", "ŋ":"ng", "ŋ-g":"ng-g", "ŋk":"nk", "ɹ":"r", "j":"y", "w":"w", "bʱ":"b", "β":"b", "ʈ͡ʂ":"ch", "ʈ͡ʂʰ":"ch","ʨ":"ch", "t͡ɕ":"ch", "t͡ɕʰ":"ch", "d̪":"d", "dʱ":"d", "ɖ":"d", "ɖʱ":"d", "ɸ":"f", "ʕ":"g", "ɡʱ":"g", "ɣ":"g", "χ":"h", "ħ":"h", "x":"h", "ɦ":"h", "d͡ʐ":"j", "d͡ʑ":"j", "ç":"h", "kʰ":"k", "q":"k", "ɫ":"l", "ɺ":"l", "ɲ":"ny", "pʰ":"p", "ʁ":"r", "ʀ":"r", "ɽ":"r", "ɽʰ":"r", "ɻ":"r", "ɾ":"r", "ʂ":"sh", "ɕ":"sh", "t̪":"t", "tʰ":"t", "ʈ":"t", "ʈʰ":"t", "ʋ":["w","v"], "r":"r", "ɰ":"w", "ɥ":"y", "ʐ":"zh", "ʑ":"zh", "ʔ":"7", "a":"a", "ä":"ah", "ɐ":"ah", "ɞ":"aw", "ɜ":"uhr", "ɒ":"oh", "o":"oh", "ʏ":"oo", "ø":"oo", "œ":"oo", "ʉ":"oo", "y":"oo", "ᵻ":"uh", "ʌ̹":"uh", "ə":"uhr", "ɘ":"uh", "ɵ":"uu", "ɯ":"uu", "ɤ":"uu"}
#ipa_nc["ʋ"] # returns ['w','v']
vowels = ['a', 'e', 'i', 'o', 'u', 'y', 'A', 'E', 'I', 'O', 'U', 'Y']
ipa_vowels = ["ɪ","ə","ɹ̩","ɨ","ɯ","ʊ","ø","ɘ","ɵ","ɤ","ɛ","œ","ɜ","ʌ","ɔ","æ","ɐ","ɑ","ɒ"]
ipa_vowels.extend(vowels)
ipa_vowels.extend([i.upper() for i in ipa_vowels])
phones = ["ah", "oh", "uh", "eh", "oo", "ee", "ie", "ou", "ae", "ai", "oi", "ei", "ue", "eu", "oe", "ay", "aa", "au", "ey", "io", "iu", "iy", "oa", "oy", "ua", "ui", "uo", "uu", "uy"]
compound_vowels = vowels + phones + [p.upper() for p in phones]

# # If modifying these scopes, delete the file token.json.
# SCOPES = ['https://www.googleapis.com/auth/documents']
# # The ID of a sample document.
# DOCUMENT_ID = '14cPRjMBaLBMUy1ZnQzr7rtko3FbH-DrEOC5YIjuhwiY'
# gdoc()

# Manual Updating and Additional Rules
ipa_nc["Ɑ"] = "AH"
ipa_nc["Ʃ"] = "S"
ipa_nc["ʃ"] = "s"
ipa_nc["ʌ"] = "e"
ipa_nc["Ʌ"] = "E"
ipa_nc["Ɒ"] = "A"
ipa_nc["ⱰL"] = "AHL"
ipa_nc["ɨ"] = "i"
ipa_nc["ɑː"] = "ah"
ipa_nc["ɒl"] = "ahl"
ipa_nc["ɒ"] = "a"
ipa_nc["aɛ"] = "ay"
ipa_nc["Æ"] = "A"
ipa_nc["Ɛ"] = "E"
ipa_nc["Ɪ"] = "I"
ipa_nc["EꞮ"] = "AY"
ipa_nc["ɪ"] = "i"
ipa_nc["⅄"] = "OO"
ipa_nc["ɒv"] = "awf"
ipa_nc["ƏƱ"] = "OH"
ipa_nc["əʊ"] = "oh"
ipa_nc["Ʊ"] = "UU"
ipa_nc["Ʋ"] = "V"
ipa_nc["ə"] = "uh" # hard remove the british english version, "uhr"
ipa_nc["ɹ̩"] = "uhr" 
ipa_nc["ɔː"] = "or"
ipa_nc["ɛː"] = "er"
# ipa_nc["əː"] = "uhr"
ipa_nc["əː"] = "ehr"
ipa_nc["eː"] = "eh"
ipa_nc["Əː"] = "EHR"
ipa_nc["ʧ"] = "ch"
ipa_nc["ɪɹ"] = "eer"
ipa_nc["dʒ"] = "j"
ipa_nc["tʃ"] = "ch" 
ipa_nc["eɪ"] = "ay"
ipa_nc["ɡ"] = "g"
ipa_nc["Ɡ"] = "G"
# now delete all identical mappings
del ipa_nc["r"]
del ipa_nc["s"]
del ipa_nc["w"]
del ipa_nc["a"]
ipa = ipa_nc.keys()
def emphasis(respelling, lang): #word to IPA with stress markers
  def listify_non_en(respelling, langcode, original):
    global vowels
    if langcode=="th": # thai case uses faster thai transliteration library
      thaipro = pronunciate(original)
      toipa = thaipro.replace("ˌ", ",").replace("ˈ", ",").replace("-", ",").strip(" ")
      thaisyll = toipa.split(",")
      transcription = [thai(s, engine="ipa") for s in thaisyll]
      print(transcription)
      return transcription
    elif langcode=="en":
      langcode="en-us"
      toipa = phonemizer.phonemize(original,backend="espeak", language=langcode, with_stress=True).strip(" ")#.replace("ˌ", ",").replace("ˈ", ",").strip(" ")
      toipa = SSP.tokenize(toipa)
      return toipa#.split(",")
    try:
      toipa = phonemizer.phonemize(original,backend="espeak", language=langcode, with_stress=True, language_switch="remove-flags").strip(" ")#.replace("ˌ", ",").replace("ˈ", ",").strip(" ")
      toipa = SSP.tokenize(toipa)
      print("NonEnglish, but Phonemizer covered", toipa)
      return toipa#.split(",")
    except RuntimeError as err:
      print(err)
      # we don't have that language covered    
      epipa = engipa.convert(respelling).strip("*") # Do we need this? emphasis/stress maybe. check with KOREAN, Arabic NAMES/SCRIPT
      epipa = SSP.tokenize(epipa)
      print("NonEnglish, using eng-to-ipa", epipa)
      return epipa
  emphasized = listify_non_en(respelling, lang, respelling)
  print(emphasized)
  print("\n")
  stress = {}
  for indx, r in enumerate(emphasized): # if this errors out, emphasized is None and we're on a language not covered. uncomment epitran section and add the language 
    if "ˈ" in r:
      stress[indx] = ("primary", r.replace("'", "").upper())
    elif "ˌ" in r:
      stress[indx] = ("secondary", r.replace("ˌ", "").upper())
    else:
      stress[indx] = ("none", r)
  mykeys = [s[0] for s in stress.values()]
  if mykeys.count("primary") > 1:
    prims = [ind for ind, p in enumerate(mykeys) if p=="primary"]
    for p in prims:
      emphasized[p] = stress[p][1]
    if "secondary" in mykeys:
      secind = mykeys.index("secondary")
      emphasized[secind] = stress[secind][1]
    for indx, e in enumerate(emphasized):
      if any([s.isupper() for s in e]) and any([s.islower() for s in e]):
        new = "".join([s.lower() for s in e])
        emphasized[indx] = new
    return "-".join(emphasized).replace("ˈ", "").replace("ˌ", "")
  if len(emphasized) == 3:
    if "primary" in mykeys:
      primind = mykeys.index("primary")
      emphasized[primind] = stress[primind][1]
    if "secondary" in mykeys:
      secind = mykeys.index("secondary")
      emphasized[secind] = stress[secind][1].replace("ˌ", "")
  elif len(emphasized) > 3:
    if "primary" in mykeys:
      primind = mykeys.index("primary")
      emphasized[primind] = stress[primind][1]
      try:
        after = emphasized[primind+1:]
        before = emphasized[:primind]
        if len(after) == 3:
          if "secondary" in mykeys[primind+1:]:
            secind = mykeys.index("secondary")
            emphasized[secind] = stress[secind][1]
        if len(before) == 3:
          if "secondary" in mykeys[:primind]:
            secind = mykeys.index("secondary")
            emphasized[secind] = stress[secind][1]
        if len(after) == 2:
          if "secondary" in mykeys[primind+1:]:
            secind = mykeys.index("secondary")
            emphasized[secind] = stress[secind][1]
        if len(before) == 2:
          if "secondary" in mykeys[:primind]:
            secind = mykeys.index("secondary")
            emphasized[secind] = stress[secind][1]
      except KeyError:
        pass
  elif len(emphasized) < 3:
    if "primary" in mykeys:
      primind = mykeys.index("primary")
      emphasized[primind] = stress[primind][1]
    if "secondary" in mykeys:
      secind = mykeys.index("secondary")
      emphasized[secind] = stress[secind][1].replace("ˌ", "")
  for indx, e in enumerate(emphasized):
    if any([s.isupper() for s in e]) and any([s.islower() for s in e]):
      new = "".join([s.lower() for s in e])
      emphasized[indx] = new
  emphout = "-".join(emphasized).replace("ˈ", "").replace("ˌ", "")
  return emphout 

def cmu_nc_map(word, lang, allo=False):
  # epipro = pre_transliterate(word)
  if not allo:
    epipro = emphasis(word, lang).strip("-") 
  else:
    epipro = word.strip("-")
  #now it's namecoach respelling time ['dajænə', 'ɹ̩ojow', 'əbɑd']
  print("Emphasis mapping outputs IPA pronunciation: ", epipro)
  # print("Mappings: ")
  map = {i:ipa_nc[i] for i in ipa if i in epipro or i.upper() in epipro}
  # print(map)
  dipthongs = ["aɪ", "eɪ", "aʊ", "eə", "ɪə", "ɔɪ",'AꞮ', 'EꞮ', 'AƱ', 'EƏ', 'ꞮƏ', 'ƆꞮ'] # removed "əʊ", 'ƏƱ'
  global vowels
  keyvowels = ['o', 'u', 'O', 'U']
  sortedmap = list(sorted(map, key=len, reverse=True)) # sort by longest key, so that you're not missing a mapping for a multicharacter IPA symbol
  # print(sortedmap)
  if "Ɑ" in sortedmap and "ɑ" in sortedmap: # put the uppercase first since it usually means A, instead of ah
    sortedmap.remove("Ɑ")
    sortedmap = ["Ɑ"] + sortedmap
  for s in sortedmap: #namecoach phonetic mapping
    val = map[s]
    # print(val)
    if s=="ʋ" or s=="Ʋ":
      if lang=="hi":
        val = "s"
      elif lang=="uk":
        val = "w"
      else:
        val = "v"
    if s in epipro: # symbol is in a non-stressed syllable
      anyd = [d for d in dipthongs if s==d]
      try:
        loc = epipro.index(s)
        if "ː" in s:
          print(s, "has a ː symbol")
          try:
            # print(loc, len(s))
            reach = epipro[loc+len(s)]
            nexto = map[reach]
            if nexto in vowels:
              nextind = sortedmap.index(reach)
              del sortedmap[nextind] # test with cases with multiple of this vowel. like in a:j -> AHy or Y, you might delete y but what if there's another y?
              epipro = epipro.replace(epipro[loc:loc+len(s)+1], nexto.upper()) # hyphenate, emphasized version
            else:
              section = epipro[loc:loc+len(s)]
              epipro = epipro.replace(section, val.upper())
          except IndexError:
            # print("No next value for: ", val.upper())
            epipro = epipro.replace(epipro[loc:loc+len(s)], val)
          except KeyError: # there's a nonmap value here
            # print("nonmap value next", reach)
            try:
              reach = epipro[loc+len(s.strip("ː"))+1]
              # print(reach, epipro)
              if reach in map:
                # print(reach, "isinmap")
                nexto = map[reach]
                if nexto in vowels:
                  # print("vowelnext")
                  nextind = sortedmap.index(reach)
                  del sortedmap[nextind] # test with cases with multiple of this vowel. like in a:j -> AHy or Y, you might delete y but what if there's another y?
                  epipro = epipro.replace(epipro[loc:loc+len(s)+1], nexto) # hyphenate, emphasized version
              else:
                # print("not in reach")
                epipro = epipro.replace(epipro[loc:loc+len(s)], val+"-")
            except IndexError:
              # print("indexerror")
              epipro = epipro.replace(epipro[loc:loc+1], val)
        elif anyd:
          if anyd==["aɪ"]:
            if epipro[:2] == "aɪ":
              val = "ay"
          print("There's at least one dipthong here! ", anyd)
          epipro = epipro.replace(epipro[loc:loc+len(s)], val)
        else:
          epipro = epipro.replace(epipro[loc:loc+len(s)], val)
          # print(epipro)
          # print(" ⬇")
      except TypeError: # the truth value of an array with more than one value... Will come up when u do a word with a nc phoneme that is a list of possible values
          print("Multiple NameCoach Phonemes Available. Picking One Randomly. ")
          epipro = epipro.replace(epipro[loc:loc+len(s)], random.choice(val, size=1)[0])
    elif s.upper() in epipro:
      anyd = [d for d in dipthongs if s==d]
      try:
        loc = epipro.index(s.upper())
        if "ː" in s:
          # print(s.upper(), "has a ː")
          # print(val)
          try:
            # print(loc, len(s))
            reach = epipro[loc+len(s)]
            nexto = map[reach]
            if nexto in vowels:
              # print("vowelnext")
              nextind = sortedmap.index(reach)
              del sortedmap[nextind] # test with cases with multiple of this vowel. like in a:j -> AHy or Y, you might delete y but what if there's another y?
              epipro = epipro.replace(epipro[loc:loc+len(s)+1], nexto.upper()) # hyphenate, emphasized version
            # else:
            #   print(epipro[loc:loc+len(s)])
            #   epipro = epipro.replace(epipro[loc:loc+len(s)], val.upper())
          except IndexError:
            # print("No next value for: ", val.upper())
            epipro = epipro.replace(epipro[loc:loc+len(s)], val.upper())
          except KeyError: # there's a hyphen here, or we're running out of characters :)
            # print("There's a hyphen here instead, KeyError")
            try:
              reach = epipro[loc+len(s.strip("ː"))+1]
              if reach in map:
                nexto = map[reach]
                if nexto in vowels:
                  nextind = sortedmap.index(reach)
                  del sortedmap[nextind] # test with cases with multiple of this vowel. like in a:j -> AHy or Y, you might delete y but what if there's another y?
                  epipro = epipro.replace(epipro[loc:loc+len(s)+1], nexto.upper()) # hyphenate, emphasized version
              else:
                epipro = epipro.replace(epipro[loc:loc+len(s)], "-"+val.upper())
            except IndexError:
              epipro = epipro.replace(epipro[loc:loc+1], val.upper())
        elif anyd:
          print("There's at least one dipthong here! CAPS...", anyd)
          if anyd==['aɪ']:
            if epipro[:2] == "AꞮ":
              val = "ay"
          epipro = epipro.replace(epipro[loc:loc+len(s)], val.upper())
        else:
          epipro = epipro.replace(epipro[loc:loc+len(s)], val.upper())
        # print(epipro)
        # print(" ⬇")
      except TypeError: # the truth value of an array with more than one value... Will come up when u do a word with a nc phoneme that is a list of possible values
          print("Multiple NameCoach Phonemes Available. Picking One Randomly. ")
          epipro = epipro.replace(epipro[loc:loc+len(s)], random.choice(val.upper(), size=1)[0])
  # print("first", epipro)
  if len(epipro)==1:
    return epipro
  elif len(epipro)==0:
    return "N/A"
  if len(epipro)>=3:
    if epipro[1] == "-" and epipro[2].isupper():# and epipro[2] not in vowels:
      epipro = epipro[0].upper() + epipro[2:]
    elif epipro[1] == "-" and epipro[2].islower() and epipro[0] in keyvowels:
      if epipro[0].isupper():
        epipro = epipro[0]+"H-" + epipro[2:]
      else:
        epipro = epipro[0]+"h-" + epipro[2:]
  tlist = epipro.split("-") # more preprocessing

  for indx, syl in enumerate(tlist):
    if "ohh" in syl:
      syl = syl.replace("ohh", "oh")
    elif "OHH" in syl:
      syl = syl.replace("OHH", "OH")
    if "ehh" in syl:
      syl.replace("ehh", "eh")
    elif "EHH" in syl:
      syl.replace("EHH", "EH")
    if "uhh" in syl:
      syl.replace("uhh", "uh")
    elif "UHH" in syl:
      syl.replace("UHH", "UH")
    if len(syl.replace("ː", "")) >= 5: # needs more splitting
      try:
        ln = detect(syl)
      except LangDetectException:
        ln = "en"
      ln = ln+"_"+ln.upper()
      try:
        pyph = pyphen.Pyphen(lang=ln)
      except KeyError:
        pyph = pyphen.Pyphen(lang="en_EN")
      hyphed = pyph.inserted(syl)
      if len(hyphed)>=5:
        pyph = pyphen.Pyphen(lang="en_EN")
        hyphed = pyph.inserted(hyphed)
      tlist[indx] = hyphed
  epipro = "-".join(tlist)
  # print(epipro)
  if epipro[0] not in vowels and epipro[1] == "-": 
    epipro = epipro[0] + epipro[2:]
  if "-" not in epipro:
    try:
      ln = detect(epipro)
    except LangDetectException:
      ln = "en"
    ln = ln+"_"+ln.upper()
    try:
      pyph = pyphen.Pyphen(lang=ln)
    except KeyError:
      pyph = pyphen.Pyphen(lang="en_EN")
    epipro = pyph.inserted(epipro)

  tlist = epipro.split("-")
  # print(tlist)
  twograms = ["uh", "ah", "eh", "oh", "ei", "ee", "oo", "ew"]
  twograms.extend([n.upper() for n in twograms])
  sylens = [3,4]
  for indx, syl in enumerate(tlist):
    sylen = len(syl)
    if indx==0 and len(tlist)>1 and sylen in sylens:
      anygram = [t for t in twograms if t in syl]
      partvl = [p for p in vowels if p in syl[1:]]
      # print(anygram[0], syl[:2])
      if anygram != [] and syl[:2]==anygram[0]:
        nexto = tlist[indx+1]
        if sylen==3:
          cut = -1
        else:
          cut = -2
        if nexto.isupper():
          tlist[indx+1] = syl[cut:].upper() + nexto
        else:
          tlist[indx+1] = syl[cut:] + nexto
        tlist[indx] = syl[:2]
      elif partvl == []:
        nexto = tlist[indx+1]
        if sylen==3:
          cut = -2
        else:
          cut = -3
        if nexto.isupper():
          tlist[indx+1] = syl[cut:].upper() + nexto
        else:
          tlist[indx+1] = syl[cut:] + nexto
        tlist[indx] = syl[0]
  # print(tlist)
  for indx, syl in enumerate(tlist):
    syl = syl.replace("ː", "")
    # print(syl)
    # If no vowels in this syllable, combine it with the previous
    if indx==0 and len(syl)>0 and any([l.isupper() for l in syl]): # if one letter is uppercase, make the whole syllable uppercase
      tlist[indx] = syl.upper()
    if indx>0: # ignore the first syllable
      if len(syl)==1 and syl not in vowels:
        curr = syl
        prev = tlist[indx-1]
        try:
          nexto = tlist[indx+1]
          # print(prev, next)
          if prev.isupper():
            tlist[indx-1] = prev + curr.upper()
            del tlist[indx]
          elif nexto.isupper():
            tlist[indx+1] = curr.upper() + nexto
            del tlist[indx]
          else:
            tlist[indx-1] = prev + curr
            del tlist[indx]
        except IndexError:
          if prev.isupper():
            tlist[indx-1] = prev + curr.upper()
            del tlist[indx]
          else:
            tlist[indx-1] = prev + curr
            del tlist[indx]
    if len(syl)>5:
      pyph = pyphen.Pyphen(lang="en_EN")
      hyphed = pyph.inserted(syl)
      tlist[indx] = hyphed
    # elif len(syl)==5 and len([x for x in "".join(syl) if x in vowels])>1:
    #   print("here")
    #   anygrams = [a for a in twograms if a in syl]
    #   print(anygrams)
    #   if anygrams != []:
    #     syl = syl.replace(anygrams[0], "-"+anygrams[0])
    #     print(syl)
    #     tlist[indx] = syl
  # print(tlist)
  for indx, syl in enumerate(tlist):
    if "`" in syl:
      tlist[indx] = syl.replace("`", "")
    if [s for s in syl if s in vowels] == [] and indx!=0:
      curr = syl
      prev = tlist[indx-1]
      # print(prev, curr)
      if prev.isupper():
        tlist[indx-1] = prev + curr.upper()
        del tlist[indx]
      else:
        tlist[indx-1] = prev + curr.lower()
        del tlist[indx]  
  # print(tlist)
  epipro = "-".join(tlist)
  epipro = epipro.replace("ː", "").replace(" ", "").replace("--", "-").strip("-")
  return epipro
  
app = Flask(__name__)

@app.route('/cmu_nc_map', methods=['GET'])
def api_cmu_nc_map():
    word = request.args.get('word', None)
    language = request.args.get('language', None)
    
    if word is None:
        return jsonify({'error': 'Word parameter are required'}), 400

    result = cmu_nc_map(word, language)
    return jsonify({'result': result})

if __name__ == '__main__':
    app.run(debug=True)